"""Top-level package for jardiff.

This module exposes the package version via the ``__version__`` attribute.
The version string follows semantic versioning.
"""

# Define the version here and use it in other modules
__all__ = ["__version__"]
__version__ = "0.2.0"
